package gkdtree;

/**
 * 
 * The Boundary class is an abstraction for n-dimensional boundaries.
 * These are used as index in an array (therefore integer).
 * In the context of Gaussian kd-trees the boundary only contains the
 * extent of the spatial structure (not the range/domain of pixels or voxels).
 * 
 * @author DarkMatt
 *
 */

public class Boundary {

	/**
	 * Lower border of the boundary.
	 */
	private int[] lower;
	/**
	 * Size of the boundary.
	 */
	private int[] size;
	/**
	 * Dimensionality of the boundary.
	 */
	private int dimension;
	
	/**
	 * 
	 * Initializes the boundary.
	 * 
	 * @param lower
	 * @param size
	 */
	public Boundary(int[] lower, int[] size) {
		this.lower = lower;
		this.size = size;
		dimension = lower.length;
	}
	
	public int[] getLower() {
		return lower;
	}
	
	public void setLower(int[] lower) {
		this.lower = lower.clone();
	}
	
	public int[] getSize() {
		return size;
	}
	
	public void setSize(int[] size) {
		this.size = size.clone();
	}
	
	/**
	 * Calculates the upper boundary from internally stored lower and size.
	 * (Calculated every time this method is executed.)
	 * 
	 * @return int[]
	 */
	public int[] getUpper() {
		int[] out = lower.clone();
		for (int i=0; i<dimension; i++) out[i] += size[i];
		return out;
	}
	
	public int getDimensionality() {
		return dimension;
	}
	
	/**
	 * Returns the maximum boundary extent. For equally sized dimensions
	 * the first dimension is returned.
	 * 
	 * @return int
	 */
	private int getMaxExtent() {
		int max = 0;
		for (int i=0;i<dimension; i++) if(size[i] > size[max]) max = i;
		return max;
	}
	
	/**
	 * Returns the midpoint of the bounardy with the highest extent.
	 * 
	 * @return int
	 */
	public int getHalfSplit() {
		int max = getMaxExtent();
		return lower[max] + size[max]/2;
	}
	
	/**
	 * A multidimensional boundary has a number of n*m*k*... elements. 
	 * Where n, m, k, ... is the size in the respective dimensions.
	 * 
	 * @return
	 */
	public int getNumberOfElements() {
		int num_elements = 1;
		for (int i=0; i<dimension; i++) num_elements *= size[i];
		return num_elements;
	}
	
	/**
	 * Splits the bounding box according to its largest dimension.
	 * The largest dimensional extent is split in the middle.
	 * This method returns both resulting boundaries.
	 * 
	 * @return Boundary[]
	 */
	public Boundary[] cutHalf() {
		int max = getMaxExtent();
		
		int[] left_size = this.size.clone();
		left_size[max] /= 2;
		Boundary left = new Boundary(this.lower.clone(), left_size);
		
		int[] right_size = left_size.clone();
		if (left_size[max] + left_size[max] != size[max]) right_size[max] ++;
		int[] right_lower = this.lower.clone();
		right_lower[max] += left_size[max];
		Boundary right = new Boundary(right_lower, right_size);
		
		return new Boundary[]{left, right};
	}
	
	/**
	 * Splits the bounding box according to a specified dimension and cut value.
	 * This method returns both resulting boundaries.
	 * 
	 * @return Boundary[]
	 */
	public Boundary[] cut(int dim, float cut) {
		
		int[] left_size = this.size.clone();
		//e.g. 4.5 -> [0..5]
		left_size[dim] = (int)Math.ceil(cut) - this.getLower()[dim];
		Boundary left = new Boundary(this.lower.clone(), left_size);
		
		//e.g. 4.5 -> [5..10]
		int[] right_lower = this.lower.clone();
		right_lower[dim] += left_size[dim];
		
		int[] right_size = this.getUpper().clone();
		for (int i=0; i<dimension; i++) right_size[i] -= right_lower[i];
		
		Boundary right = new Boundary(right_lower, right_size);
		
		return new Boundary[]{left, right};
	}
	
	/**
	 * Returns true if a point at integer coordinates is inside the boundary.
	 * 
	 * @param point
	 * @return boolean
	 */
	public boolean inside(int[] point) {
		int[] upper = getUpper();
		int counter = 0;
		for (int i=0; i<dimension; i++) {
			if (point[i] >= lower[i] && point[i] < upper[i]) {
				counter++;
			}
		}
		if (counter == dimension) return true;
		return false;
	}
	
	/**
	 * Returns true if the bounding boxes are not overlapping.
	 * This is an n-dimensional to n-dimensional intersection test.
	 * (It is assumed that the dimensions match.)
	 * 
	 * @param b
	 * @return boolean
	 */
	public boolean non_overlapping(Boundary b) {
		for (int i=0; i<dimension; i++) {
			if (lower[i] > b.getUpper()[i]) {
				return true;
			}
			if (getUpper()[i] < b.getLower()[i]) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * True if the current boundary is totally contained within b.
	 * (It is assumed that the dimensions match.)
	 * 
	 * @param b
	 * @return boolean
	 */
	public boolean containedBy(Boundary b) {
		for (int i=0; i<dimension; i++) {
			if (lower[i] < b.getLower()[i]) {
				return false;
			}
			if (getUpper()[i] > b.getUpper()[i]) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * For a given boundary, calculate the absolute offset of a 
	 * boundary contained within this boundary.
	 * 
	 * @param subregion
	 * @return Boundary
	 */
	public Boundary getAbsoluteSubregion(Boundary subregion) {
		int[] l = new int[dimension];
		for (int i=0; i<dimension; i++) {
			l[i] = subregion.getLower()[i] - lower[i];
		}
		return new Boundary(l, subregion.getSize());
	}
	
	/**
	 * Returns the minimal sub-region of two boundaries.
	 * (the largest lower and the smallest higher)
	 * 
	 * @param region
	 * @return Boundary
	 */
	public Boundary getMinimalSubregion(Boundary region) {
		//System.out.println("getMinimalSubregion");
		//this.printBoundary("this");
		//region.printBoundary("b");
		
		// lower
		int[] l = region.getLower().clone();
		for (int i=0; i<dimension; i++) {
			if (region.getLower()[i] < getLower()[i])
				l[i] = getLower()[i];
		}
		// size
		int[] s = region.getSize().clone();
		for (int i=0; i<dimension; i++) {
			if (region.getUpper()[i] > getUpper()[i]) {
				s[i] = getUpper()[i]-l[i];
			} else {
				s[i] = region.getUpper()[i]-l[i];
			}
		}
		
		return new Boundary(l, s);
	}
	
	/**
	 * Prints the boundary information to the console.
	 * 
	 * @param id
	 */
	public void printBoundary(String id) {
		System.out.print("Boundary ("+id+") lower/upper: ");
		int[] upper = getUpper();
		for (int i=0; i<dimension-1; i++) {
			System.out.print(lower[i]+"/"+upper[i]+", ");
		}
		System.out.println(lower[dimension-1]+"/"+upper[dimension-1]+".");
	}
	
}
