package gkdtree;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.reflect.Array;
import java.net.URI;
import java.net.URISyntaxException;


/**
 * 
 * The File-Node class handles file save and load operations. If a file does not fit into the main memory, 
 * the file load is handled in subsequent load steps.
 * 
 * The File-Node is part of a Quad/Octree structure if the data is too large to fit into main memory.
 * 
 * The generic parameter T is for the data values (e.g. every point inside the data is of type Short).
 * 
 * Storage data-structure:
 * 
 *  - File: header: 
 *     - dimension d
 *     - d * size (only valid for original file)
 *  - Corresponding File-Class: 
 *     - file-name
 *     - lower/upper bounds
 *     - size
 * 
 * General comments:
 *  - Loads a single file and its partitions
 *  - Sometimes only a subset of points contained within a FileNode is required. In case of selecting
 *  a boundary inside the spatial dimension with limited domain (e.g. only consider values between 1 and 10).
 *  - (A file-node is a dimension smaller if one value is stored at every data point.)
 *  
 *  Ideas during implementation:
 *   - blurring a vector-field
 * 
 * @author DarkMatt
 * @param <T>
 *
 */
public class FileNode<T> {

	private static final boolean DEBUG = true;
	/**
	 * The maximum size that can fit into main memory.
	 * (needs to be: data-size = i * MAX_ARRAY_SIZE)
	 */
	public static final int MAX_ARRAY_SIZE = 125000;//10000;//25;//1*100*100;
	
	/**
	 * The full file-name.
	 */
	public String filename;
	/**
	 * Boundaries of the data. When loading a file, the original
	 */
	public Boundary b;
	/**
	 * Left child of the file-octree structure.
	 */
	public int left_child;
	/**
	 * Right child of the file-octree structure.
	 */
	public int right_child;
	/**
	 * Dimensionality of the data in its spatial extent.
	 * This is the overall dimensionality - 1.
	 * (If one value is stored at every grid point.)
	 */
	public int dimension;
	/**
	 * Number of partitions required to read the file.
	 */
	public int partition_count;
	/**
	 * The current part of the file loaded.
	 * This is reset on init-load.
	 */
	public int current_file_part_load;
	private int number_of_values = 0;
	public int[] offset;
	
	/**
	 * For determining the type of T.
	 */
	private Class<T> type;
	/**
	 * Read-buffer is the maximal size of the data in memory.
	 */
	private static int readBuffer = MAX_ARRAY_SIZE;
	
	private File file;
	//load
	private FileInputStream fileIs;
	private ObjectInputStream is;
	private boolean load_first_element;
	private T[] data;
	//write
	private FileOutputStream fileOs;
	private ObjectOutputStream os;
	
	private File absolute_path;
	
	/**
	 * A file-node is initialized with a relative but complete path name.
	 * The level (depth) inside the tree and an index going from the left most
	 * element to the right one.
	 * A boundary defines the extent of the data and the type of the data is defined
	 * with the type parameter.
	 * 
	 * @param filename
	 * @param level
	 * @param index
	 * @param b
	 * @param type
	 */
	public FileNode(String filename, int level, int index, Boundary b, Class<T> type) {
		this.filename = filename;
		this.type = type;
		this.b = b;
		
		left_child = -1;
		right_child = -1;
		
		//System.out.println("=================== LEVEL "+level+" ======== INDEX "+index+" ===================");
		
		//b.printBoundary(level+"/"+index);
		
		//copy
		offset = b.getLower().clone();
		
		//MAX_ARRAY_SIZE = b.getSize()[0] *  b.getSize()[1];
		
		
		//absolute_path
		
		URI uri = null;
		try {
			uri = new URI("file://"+FileNode.class.getProtectionDomain().getCodeSource().getLocation().getPath());
		} catch (URISyntaxException e) {
			e.printStackTrace();
		}
		absolute_path = new File(new File(uri).getParent());
		
	}
	
	public static void setReadBufferSize(int in) {
		readBuffer = in;
	}
	
	/**
	 * Method to create an empty file with the specified filename.
	 */
	public void createFile() {
		
		File f = new File(absolute_path +"/"+ filename);
		if (!f.exists()) {
			try {
				f.createNewFile();
			} catch (IOException e) {
				System.out.println(f.getAbsolutePath());
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Initializes a file-write.
	 * After initializing a file write, multiple data-block can be written to the opened file.
	 * (The final size is unknown at this point.)
	 * 
	 * @param data_size
	 */
	public void initFileWrite(int[] data_size) {
		number_of_values = -1; //TODO why???
		dimension = data_size.length;
		
		file = new File(absolute_path +"/"+ filename);
		if (file.exists()) {
		
			try {
				fileOs = new FileOutputStream(absolute_path +"/"+ filename);
				os = new ObjectOutputStream(new BufferedOutputStream(fileOs));
				
				os.writeInt(dimension);
				for (int i=0; i<dimension; i++) {
					os.writeInt(data_size[i]);
				}
				
			} catch (IOException e) {
				e.printStackTrace();
			}
		
		} else {
			System.out.println("ERROR: file ("+filename+") does not exist");
		}	
		
	}
	
	/**
	 * Writes a data block to the opened file.
	 * (If there is no opened file an exception is thrown.)
	 * 
	 * @param data
	 * @param size
	 */
	public void writeNextFilePart(short[] data, int size) {
		
		try {
			
			for (int i=0; i < size; i++) {
				
				os.writeShort(data[i]);
				
			}
			
			number_of_values += size;
			
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	
	/**
	 * Writes a single element to the currently opened file.
	 * 
	 * @param data
	 */
	public void writeNextFilePart(short data) {
		try {
			os.writeShort(data);
			number_of_values ++;
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Closes the currently opened file, which was used for
	 * a file write.
	 */
	public void closeWriteFile() {
		
		try {
			os.close();
			fileOs.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	
	/**
	 * Initializes a file load.
	 * Reads the data size and allows successive loading of file chunks.
	 */
	@SuppressWarnings({"unchecked"})
	public void initFileLoad() {
		
		file = new File(absolute_path +"/"+ filename);
		if (file.exists()) {
			
			try {
				fileIs = new FileInputStream(absolute_path +"/"+ filename);
				is = new ObjectInputStream(new BufferedInputStream(fileIs));
				
				dimension = is.readInt();
				
				//System.out.println("dim: "+dimension);
				
				int[] tmp_size = new int[dimension];
				for (int i=0; i<dimension; i++) {
					tmp_size[i] = is.readInt();
				}
				b.setSize(tmp_size);
				
				partition_count = b.getNumberOfElements()/readBuffer;
				
				data = (T[]) Array.newInstance(type, readBuffer);
				
				if (DEBUG)
					System.out.println("Initialized file-load ("+filename+"), size: " 
							+ b.getNumberOfElements() +".");
				
			} catch (IOException e) {
				System.out.println(absolute_path +"/"+ filename);
				e.printStackTrace();
			}
			
			load_first_element = true;
			current_file_part_load = 0;
			offset = b.getLower().clone();
			
		} else {
			System.out.println("ERROR: file ("+filename+") does not exist");
		}
	}
	
	/**
	 * Loads an array of data of type T from the currently opened file.
	 * In addition the offset parameter defines the current starting point
	 * inside the array as an n-dimensional position.
	 * 
	 * The size of the output is always the maximal size, except for the last
	 * read which may be smaller.
	 * 
	 * @return T[]
	 */
	@SuppressWarnings({"unchecked"})
	public T[] loadNextFilePart() {
		
		if (file.exists()) {
			
			try {
				
				data = (T[]) Array.newInstance(type, readBuffer);
				
				current_file_part_load = 0;
				
				for (int i=0; i < readBuffer; i++) {
					if (current_file_part_load < b.getNumberOfElements())
						if (type == Short.class) {
							
							//System.out.println("i am doing what i am supposed to do");
							
							data[i] = (T) (Short)is.readShort(); //TODO very ugly!
							//System.out.println(data[i]);
						}
					else
						System.out.println("Error: Reading out of File.");
					current_file_part_load ++;
				}
				
				if (!load_first_element) {
					for (int i=0; i<dimension; i++) {
						offset[i] += LinearNDMapping.toNDimensional(dimension, b.getSize(), readBuffer)[i];
					}
				} else {
					load_first_element = false;
				}
				
			} catch (Exception e) {
				System.out.println(current_file_part_load);
				e.printStackTrace();
			}
			
		}
		
		return data;
	}
	
	/**
	 * Closes the current file load.
	 */
	public void closeLoadFile() {
		try {
			is.close();
			fileIs.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		data = null;
	}
	
	/**
	 * Reads a single file. This is only possible for leaf nodes.
	 * Otherwise the data may not fit into memory.
	 * 
	 * @return T[]
	 */
	@SuppressWarnings({"unchecked"})
	public T[] getSingleFile() {
		
		//System.out.println("Init single file load...");
		
		file = new File(absolute_path +"/"+ filename);
		if (file.exists()) {
			try {
				fileIs = new FileInputStream(absolute_path +"/"+ filename);
				is = new ObjectInputStream(new BufferedInputStream(fileIs));
				
				dimension = is.readInt();
				int[] tmp_size = new int[dimension];
				for (int i=0; i<dimension; i++) {
					tmp_size[i] = is.readInt();
				}
				b.setSize(tmp_size);
				
				/*if (DEBUG) {
					for (int i=0; i<tmp_size.length; i++) {
						System.out.println("dimension " + i + ":" + tmp_size[i]);
					}
				}*/
				
				// new // readBuffer default MAX_ARRAY_SIZE
				data = (T[]) Array.newInstance(type, b.getNumberOfElements());
				
				/*if (DEBUG) {
					System.out.println("Initialized single file-load ("+filename+"), size: " 
							+ b.getNumberOfElements() +".");
				}*/
				
				// fill
				current_file_part_load = 0;
				
				for (int i=0; i < readBuffer; i++) {
					//if (DEBUG) System.out.println(i);
					if (current_file_part_load < b.getNumberOfElements())
						if (type == Short.class) {
							data[i] = (T) (Short)is.readShort();
						}
					else
						System.out.println("Error: Reading out of File.");
					current_file_part_load ++;
				}
				
				is.close();
				fileIs.close();
				
			} catch (IOException e) {
				System.out.println(absolute_path +"/"+ filename);
				e.printStackTrace();
			}
			
		} else {
			System.out.println("ERROR: file ("+filename+") does not exist");
		}
		
		if (DEBUG) {
			for (int i=0; i<data.length; i++) {
				if (data[i] == null) System.out.println(data[i]);
			}
		}
		
		//return null;
		return data;
	}
	
	/**
	 * Sets a leaf node to the data input array.
	 * 
	 * @param data
	 */
	public void setSingleFile(T[] data) {
		
		//System.out.println("Init single file load...");
		
		file = new File(absolute_path +"/"+ filename);
		if (file.exists()) {
			try {
				fileOs = new FileOutputStream(absolute_path +"/"+ filename);
				os = new ObjectOutputStream(new BufferedOutputStream(fileOs));
				
				os.writeInt(dimension);
				int[] size = b.getSize();
				for (int i=0; i<dimension; i++) {
					os.writeInt(size[i]);
				}
				
				// TODO: verify
				for (int i=0; i < data.length; i++) {
					os.writeShort((Short)data[i]);
				}
				
				os.close();
				fileOs.close();
			} catch (IOException e) {
				System.out.println(absolute_path +"/"+ filename);
				e.printStackTrace();
			}
		} else {
			System.out.println("ERROR: file ("+filename+") does not exist");
		}
		
		if (DEBUG) {
			for (int i=0; i<data.length; i++) {
				if (data[i] == null) System.out.println(data[i]);
			}
		}
		
	}
	
	
	public int get_number_of_values() {
		return number_of_values + 1;
	}
	
	
}
