package gkdtree;

import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;

import javax.imageio.ImageIO;

/**
 *
 * A quad/octree structure for files. Depending on the file size the file
 * is partitioned into subfiles. Subfiles can be read fully into memory.
 * 
 * @author DarkMatt
 *
 */

public class FileOctree {

	/**
	 * needs to be: I * MAX_NODE_SIZE = MAX_ARRAY_SIZE
	 */
	private static final int MAX_NODE_SIZE = 25000;//2500;//25;//1*50*50;
	private static final boolean DEBUG = true;
	
	// Tree parameters
	private ArrayList<FileNode<Short>> nodes;
	private HashSet<Integer> leaf_indices;
	
	// Query parameters
	private ArrayList<Integer> query_nodes; 
	
	private static int maximum_size = MAX_NODE_SIZE;
	
	// Get List Parameters
	private Boundary selection;
	private int currentPartition;
	private int[] partitionsIndices;
	
	public static void setMaximumSize(int in) {
		maximum_size = in;
	}
	
	/**
	 * Initializes the file and creates the octree.
	 * 
	 * @param max_partition_size
	 * @param path
	 * @return boolean
	 */
	public boolean initFile(int max_partition_size, String path) {
		
		URI uri = null;
		try {
			uri = new URI("file://"+FileOctree.class.getProtectionDomain().getCodeSource().getLocation().getPath());
		} catch (URISyntaxException e) {
			e.printStackTrace();
		}
		File absolute_path = new File(new File(uri).getParent());
		
		if (max_partition_size > 0) maximum_size = max_partition_size;
		File f = new File(absolute_path + "/" + path);
		
		if (f.exists()) {
			
			nodes = new ArrayList<FileNode<Short>>();
			leaf_indices = new HashSet<Integer>();
			
			int[] lower = null;
			int[] size = null;
			
			try {
				FileInputStream fileIs = new FileInputStream(path);
				ObjectInputStream is = new ObjectInputStream(new BufferedInputStream(fileIs));
				
				// read header
				int dim = is.readInt();
				lower = new int[dim];
				size = new int[dim];
				for (int d=0; d<dim; d++) {
					lower[d] = 0;
					size[d] = is.readInt();
				}
				
				is.close();
				fileIs.close();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
			
			//left-right is arbitrary here
			addNode(path, true, -1, 0, 0, new Boundary(lower, size));
			
			if (DEBUG) {
				/*System.out.println("Final tree size: " + nodes.size());
				System.out.println("Final leaf nodes: ");
				for (int i=0; i<leaf_indices.size()-1; i++) 
					System.out.print(leaf_indices.toArray()[i] + ", ");
				System.out.println(leaf_indices.toArray()[leaf_indices.size()-1]);*/
			}
			
			writeLeafFiles();
			
		} else {
			return false;
		}
		
		return true;
		
	}
	
	/**
	 * The recursive method to create the octree.
	 * 
	 * The name of the node consists of:
	 *  filename_level_index
	 *  e.g. testdata_1_0 or testdata_2_0
	 *  for the first element in the second and third level.
	 * 
	 * @param name the file-name without suffix
	 * @param left true if left child, false if right child
	 * @param previous_node index to the previous node in the node list
	 * @param level level (depth) of the node in the tree structure
	 * @param index an index at the current level
	 * @param b
	 */
	private void addNode(String name, boolean left, int previous_node, int level, int index, Boundary b) {
		
		if (b.getNumberOfElements() < maximum_size) {
			//if (DEBUG) b.printBoundary(level+"/"+index);
			//if (DEBUG) System.out.println("Previous: " + previous_node);
			leaf_indices.add(previous_node);
			return;
		}
		
		FileNode<Short> node = new FileNode<Short>(name+"_"+level+"_"+index, level, index, b, Short.class);
		nodes.add(node);
		
		if (previous_node != -1) {
			if (left) nodes.get(previous_node).left_child = nodes.size()-1;
			if (!left) nodes.get(previous_node).right_child = nodes.size()-1;
		}
		
		previous_node = nodes.size()-1;
		
		Boundary[] newBoundaries = b.cutHalf();
		addNode(name, true, previous_node, level+1, 2*index, newBoundaries[0]);
		addNode(name, false, previous_node, level+1, 2*index+1, newBoundaries[1]);
	}
	
	/**
	 * Write data to the generated leaf files. These are based on the pre-generated
	 * octree nodes. Which are initialized empty.
	 * 
	 * This method writes the leaf files information to a file depending on the 
	 * calculated boundaries.
	 */
	private void writeLeafFiles() {
		
		Object[] tmp = leaf_indices.toArray();
		Integer[] leaves = Arrays.copyOf(tmp, tmp.length, Integer[].class);
		
		//init root node read
		FileNode<Short> root = nodes.get(0);
		root.initFileLoad();
		
		//init leaf nodes
		for(int i=0; i<leaves.length; i++) {
			FileNode<Short> cur = nodes.get(leaves[i]);
			cur.createFile();
			
			//System.out.println(cur.b.getSize()[0] + "/" + cur.b.getSize()[1] + "/" + cur.b.getSize()[2]);
			
			cur.initFileWrite(cur.b.getSize());
		}
		
		for (int p=0; p<root.partition_count; p++) {
		
			Short[] data = root.loadNextFilePart();
			
			for (int i=0; i<data.length; i++) {
				
				int pos[] = LinearNDMapping.toNDimensional(
						root.dimension, root.b.getSize(), i);
				for (int off=0; off<root.dimension; off++) {
					pos[off] += root.offset[off];
				}
				
				//System.out.println(pos[0]+"/"+pos[1]+"/"+pos[2]);
				for(int j=0; j<leaves.length; j++) {
					FileNode<Short> cur = nodes.get(leaves[j]);
					if (cur.b.inside(pos)) {
						cur.writeNextFilePart(data[i]);
					}
				}
				
			}
		
		}
		
		for(int i=0; i<leaves.length; i++) {
			FileNode<Short> cur = nodes.get(leaves[i]);
			cur.closeWriteFile();
		}
		root.closeLoadFile();
		
		// ERROR CHECKING!
		/*for(int j=0; j<leaves.length; j++) {
			FileNode<Short> cur = nodes.get(leaves[j]);
			cur.initFileLoad();
			Short[] check = cur.loadNextFilePart();
			for (int i=0; i<cur.number_of_values; i++) {
				if (check[i] <= 0) System.out.println("error");
			}
			cur.closeLoadFile();
		}*/
		
	}
	
	/**
	 * Finds the leaf nodes contained within a certain area.
	 * It returns the indices of the nodes.
	 * 
	 * The overview of the approach:
	 *  1. Query the tree area.
	 *  2. Only get the contained leaves.
	 * 
	 * The idea of the query (recursion):
	 *  - If Boundary.non_overlapping() the query area and the octree cell are
	 *  disconnected (non overlapping), therefore the recursion can stop. 
	 *  - If the tree leaf is fully contained within the query area, 
	 *  the all the leaves in this branch can be added. No more recursion is necessary.
	 *  - Otherwise keep going through the tree and adding the the nodes. (recursion)
	 * 
	 * (implemented in queryRecursion and addChilds)
	 * 
	 * @param area
	 * @return int[]
	 */
	public int[] queryArea(Boundary area) {
		
		query_nodes = new ArrayList<Integer>();
		
		queryRecursion(0, area);
		
		Object[] tmp = leaf_indices.toArray();
		Integer[] leaves = Arrays.copyOf(tmp, tmp.length, Integer[].class);
		
		// remove all indices that are not leaves!
		for (int i=0; i<query_nodes.size(); i++) {
			boolean found = false;
			for (int j=0; j<leaves.length; j++) {
				if (query_nodes.get(i) == leaves[j]) found = true; //ERROR WAS HERE!1
			}
			if (!found) {
				query_nodes.remove(i);
				i--;
			}
		}
		
		// write to array
		int[] out = new int[query_nodes.size()];
		for (int i=0; i<query_nodes.size(); i++) {
			out[i] = query_nodes.get(i);
		}
		
		if (DEBUG) {
			//System.out.println("Query with " + out.length + " elements.");
			//area.printBoundary("Query area");
		}
		
		return out;
	}
	
	/**
	 * The recursion of the queryArea method.
	 * 
	 * @param currentnode
	 * @param area
	 */
	private void queryRecursion(int currentnode, Boundary area) {
		
		//System.out.println(currentnode);
		
		if (currentnode == -1) return;
		
		if (nodes.get(currentnode).b.non_overlapping(area)) {
			/*System.out.println("Non-overlapping: " + currentnode);
			System.out.println(
					nodes.get(currentnode).b.getLower()[0] + "/" +
					nodes.get(currentnode).b.getLower()[1] + "/" +
					nodes.get(currentnode).b.getLower()[2]);
			System.out.println(
					nodes.get(currentnode).b.getSize()[0] + "/" +
					nodes.get(currentnode).b.getSize()[1] + "/" +
					nodes.get(currentnode).b.getSize()[2]);*/
			return;
		}
		
		if (nodes.get(currentnode).b.containedBy(area)) {
			//take all
			addChilds(currentnode);
		} else {
			//split
			query_nodes.add(currentnode);
			queryRecursion(nodes.get(currentnode).left_child,area);
			queryRecursion(nodes.get(currentnode).right_child,area);
		}
		
	}
	
	/**
	 * Adds all the children on a node in a recursion.
	 * 
	 * @param current
	 */
	private void addChilds(int current) {
		if (current != -1) {
			query_nodes.add(current);
		} else {
			return;
		}
		
		addChilds(nodes.get(current).left_child);
		addChilds(nodes.get(current).right_child);
	}
	
	/**
	 * Returns the boundary of a certain node.
	 * 
	 * @param element
	 * @return Boundary
	 */
	public Boundary getBoundaryOfElement(int element) {
		return nodes.get(element).b;
	}
	
	/**
	 * Gets the short values of a certain leaf node.
	 * 
	 * @param element
	 * @return Short[]
	 */
	public Short[] getValues(int element) {
		nodes.get(element).initFileLoad();
		Short[] data = nodes.get(element).loadNextFilePart();
		nodes.get(element).closeLoadFile();
		
		return data;
	}
	
	/**
	 * Generates a random data-set.
	 * The sphere boolean only works in 3d and generates a sphere point cloud.
	 * 
	 * @param filename
	 * @param size
	 * @param sphere
	 */
	public void generateNewShort(String filename, int[] size, boolean sphere) {
		
		Random r = new Random();
		
		try {
			FileOutputStream fileOs = new FileOutputStream(filename+"_0_0");
			ObjectOutputStream os = new ObjectOutputStream(new BufferedOutputStream(fileOs));
			
			
			FileOutputStream fileOs_second = new FileOutputStream(filename);
			ObjectOutputStream os_second = new ObjectOutputStream(new BufferedOutputStream(fileOs_second));
			
			
			int total_size = 1;
			
			// write header
			os.writeInt(size.length);
			os_second.writeInt(size.length);
			for (int i=0; i<size.length; i++) {
				total_size *= size[i];
				os.writeInt(size[i]);
				os_second.writeInt(size[i]);
			}
			
			for (int i=0; i<total_size; i++) {
				//os.writeFloat(r.nextFloat());
				//os.writeByte((byte) r.nextInt(Byte.MAX_VALUE + 1));
				
				//short val = (short) r.nextInt(Short.MAX_VALUE + 1);
				
				int rand = r.nextInt(10000);
				short val = (short) (Short.MAX_VALUE/2 - rand);
				
				if (sphere) {
					int[] pos = LinearNDMapping.toNDimensional(3, new int[]{100,100,100},i);
					
					if (Math.pow(pos[0]-50,2) + Math.pow(pos[1]-50,2) + Math.pow(pos[2]-50,2) > 2400) {
						val = 0;
					} else if (Math.pow(pos[0]-50,2) + Math.pow(pos[1]-50,2) + Math.pow(pos[2]-50,2) < 2300) {
						val = 0;
					}
				}
				
				os.writeShort(val);
				os_second.writeShort(val);
				
			}
			
			os.close();
			os_second.close();
			fileOs.close();
			fileOs_second.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	
	/**
	 * Generates a 2d checker data-set.
	 * 
	 * @param filename
	 * @param size
	 */
	public void generateNew2DChecker(String filename, int[] size) {
		
		Random r = new Random();
		
		try {
			FileOutputStream fileOs = new FileOutputStream(filename+"_0_0");
			ObjectOutputStream os = new ObjectOutputStream(new BufferedOutputStream(fileOs));
			
			
			FileOutputStream fileOs_second = new FileOutputStream(filename);
			ObjectOutputStream os_second = new ObjectOutputStream(new BufferedOutputStream(fileOs_second));
			
			
			int total_size = 1;
			
			// write header
			os.writeInt(size.length);
			os_second.writeInt(size.length);
			for (int i=0; i<size.length; i++) {
				total_size *= size[i];
				os.writeInt(size[i]);
				os_second.writeInt(size[i]);
			}
			
			for (int i=0; i<total_size; i++) {
				//os.writeFloat(r.nextFloat());
				//os.writeByte((byte) r.nextInt(Byte.MAX_VALUE + 1));
				
				//short val = (short) r.nextInt(Short.MAX_VALUE + 1);
				
				int rand = r.nextInt(3000);
				
				int[] pos = LinearNDMapping.toNDimensional(2, size,i);
				
				short val;
				if ((pos[0] < size[0]/2 && pos[1] < size[0]/2) 
						|| (pos[0] > size[0]/2-1 && pos[1] > size[0]/2-1)) {
					val = (short) (3*Short.MAX_VALUE/4 - rand);
				} else {
					val = (short) (Short.MAX_VALUE/4 - rand);
				}
				
				/*if (sphere) {
					int[] pos = LinearNDMapping.toNDimensional(3, new int[]{100,100,100},i);
					
					if (Math.pow(pos[0]-50,2) + Math.pow(pos[1]-50,2) + Math.pow(pos[2]-50,2) > 2400) {
						val = 0;
					} else if (Math.pow(pos[0]-50,2) + Math.pow(pos[1]-50,2) + Math.pow(pos[2]-50,2) < 2300) {
						val = 0;
					}
				}*/
				
				os.writeShort(val);
				os_second.writeShort(val);
				
			}
			
			os.close();
			os_second.close();
			fileOs.close();
			fileOs_second.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	
	/**
	 * Generates a new data-set from a png image.
	 * 
	 * @param filename
	 * @param source_filename
	 */
	public void generateNew2DFile(String filename, String source_filename, int[] size) {
		
		URI uri = null;
		try {
			uri = new URI("file://"+FileOctree.class.getProtectionDomain().getCodeSource().getLocation().getPath());
		} catch (URISyntaxException e) {
			e.printStackTrace();
		}
		File absolute_path = new File(new File(uri).getParent());
		
		//load from file...
		BufferedImage img = null;
		try {
			img = ImageIO.read(new File(absolute_path+"/"+source_filename));
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		
		DataBufferByte data_buffer = (DataBufferByte)img.getRaster().getDataBuffer();
		byte[] data = data_buffer.getData(); //RGBA?
		
		System.out.println("Image size: " + data.length + " s: " + size[0] + "/" + size[1]);
		
		int lin_size = 1;
		for (int i=0; i<size.length; i++) lin_size *= size[i];
		FileNode.setReadBufferSize(lin_size/2);
		FileOctree.setMaximumSize(lin_size/4);
		
		//int[] size = new int[]{100, 100}; //removed
		
		try {
			FileOutputStream fileOs = new FileOutputStream(filename+"_0_0");
			ObjectOutputStream os = new ObjectOutputStream(new BufferedOutputStream(fileOs));
			
			
			FileOutputStream fileOs_second = new FileOutputStream(filename);
			ObjectOutputStream os_second = new ObjectOutputStream(new BufferedOutputStream(fileOs_second));
			
			
			int total_size = 1;
			
			// write header
			os.writeInt(size.length);
			os_second.writeInt(size.length);
			for (int i=0; i<size.length; i++) {
				total_size *= size[i];
				os.writeInt(size[i]);
				os_second.writeInt(size[i]);
			}
			
			for (int i=0; i<total_size; i++) {
				
				short val = (short)((int)data[4*i+1] & 0xff); //accesses R value?
				
				// MAP: 0-255 to 1-Short.Maxvalue
				//TODO: unintuitive scaling?
				val += 1; //4
				val *= Short.MAX_VALUE/256; //260
				
				//scale
				//val *= Short.MAX_VALUE/256;
				
				//if (data[4*i] == data[4*i+1] && data[4*i+2] == data[4*i+1]) System.out.println("OK..................................");
				
				//if ((int)data[4*i+1] != -1) System.out.println("Data: " + (data[4*i+1]));
				
				//System.out.println(val);
				
				os.writeShort(val);
				os_second.writeShort(val);
				
			}
			
			os.close();
			os_second.close();
			fileOs.close();
			fileOs_second.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	
	/**
	 * Returns an array of points for rendering.
	 * The points are 3d positions which are intended to be 
	 * extended to quads in a geometry shader.
	 * 
	 * @param threshold
	 * @return float[]
	 */
	public float[] getPoints(int threshold) {
		
		FileNode<Short> root = nodes.get(0);
		//int m = root.dimension + 1;
		//new float[m * root.b.getNumberOfElements()]
		ArrayList<Float> out = new ArrayList<Float>();
		
		root.initFileLoad();
		//int c = 0;
		for (int i=0; i<root.partition_count; i++) {
			
			Short[] data = (Short[])root.loadNextFilePart();
			
			for (int j=0; j<root.current_file_part_load; j++) {
				
				int[] pos = LinearNDMapping.toNDimensional(root.dimension, root.b.getSize(), j);
				for (int off=0; off<root.dimension; off++) {
					pos[off] += root.offset[off];
				}
				
				if (data[j] > threshold) {
					for (int x=0; x<root.dimension; x++) {
						//out[m*c+x] = (float)pos[x]/(float)root.b.getSize()[x];
						out.add((float)pos[x]/(float)root.b.getSize()[x]);
					}
					//out[m*c+root.dimension] = (float)data[j]/(float)Short.MAX_VALUE;
					out.add((float)data[j]/(float)Short.MAX_VALUE);	
				} else {
					//System.out.println(data[j]);
				}
				
				//c ++;
			}
		}
		
		root.closeLoadFile();
		
		float[] out_array = new float[out.size()];
		
		for (int i=0; i<out_array.length; i++) {
			out_array[i] = out.get(i);
		}
		
		return out_array;
	}
	
	/**
	 * Returns an area selection of the octree as a list of points.
	 * 
	 * @param threshold
	 * @param active_parts
	 * @return float[]
	 */
	public float[] getPointsOctree(int threshold, int[] active_parts) {
		
		Object[] tmp = leaf_indices.toArray();
		Integer[] leaves = Arrays.copyOf(tmp, tmp.length, Integer[].class);
		
		ArrayList<Float> out = new ArrayList<Float>();
		
		for (int k=0; k<leaves.length; k++) {
		
			boolean execute = false;
			for (int j=0; j<active_parts.length; j++) {
				if (leaves[k] == active_parts[j]) execute = true;
			}
			
			if (execute) {
				FileNode<Short> curr = nodes.get(leaves[k]);
				//new float[m * root.b.getNumberOfElements()]
				
				curr.initFileLoad();
				for (int i=0; i<1; i++) {
					
					Short[] data = (Short[])curr.loadNextFilePart();
					
					//System.out.println("num_val: " + root.number_of_values);
					for (int j=0; j<curr.get_number_of_values(); j++) {
						
						int[] pos = LinearNDMapping.toNDimensional(curr.dimension, curr.b.getSize(), j);
						for (int off=0; off<curr.dimension; off++) {
							pos[off] += curr.offset[off];
						}
						
						if (data[j] > threshold) {
							for (int x=0; x<curr.dimension; x++) {
								out.add((float)pos[x]/(float)100);//root.b.getSize()[x]);
							}
							out.add((float)data[j]/(float)Short.MAX_VALUE);
						}
						
					}
				}
				
				curr.closeLoadFile();
			}
		
		}
		
		float[] out_array = new float[out.size()];
		
		for (int i=0; i<out_array.length; i++) {
			out_array[i] = out.get(i);
		}
		
		return out_array;
	}
	
	/**
	 * Most recent implementation to get a list of points from a 2D image.
	 * This is a 2D height-map. (Currently used)
	 * 
	 * @param threshold
	 * @param active_parts
	 * @return float[]
	 */
	public float[] getPointsOctree2D(int threshold, int[] active_parts) {
		
		FileNode<Short> root = nodes.get(0);
		
		Object[] tmp = leaf_indices.toArray();
		Integer[] leaves = Arrays.copyOf(tmp, tmp.length, Integer[].class);
		
		ArrayList<Float> out = new ArrayList<Float>();
		
		for (int k=0; k<leaves.length; k++) {
		
			boolean execute = false;
			for (int j=0; j<active_parts.length; j++) {
				if (leaves[k] == active_parts[j]) execute = true;
			}
			
			if (execute) {
				FileNode<Short> curr = nodes.get(leaves[k]);
				//new float[m * root.b.getNumberOfElements()]
				
				curr.initFileLoad();
				//for (int i=0; i<1; i++) {
					
					Short[] data = (Short[])curr.loadNextFilePart();
					
					//System.out.println("num_val: " + root.number_of_values);
					for (int j=0; j<curr.get_number_of_values(); j++) {
						
						int[] pos = LinearNDMapping.toNDimensional(curr.dimension, curr.b.getSize(), j);
						for (int off=0; off<curr.dimension; off++) {
							pos[off] += curr.offset[off];
						}
						
						if (data[j] > threshold) {
							for (int x=0; x<curr.dimension; x++) {
								out.add((float)(pos[x]+0.5f)/root.b.getSize()[x]);//(float)100);//root.b.getSize()[x]);
							}
							out.add((float)data[j]/((float)Short.MAX_VALUE)*1.0f);
						}
						
					}
				//}
				
				curr.closeLoadFile();
			}
		
		}
		
		float[] out_array = new float[out.size()];
		
		for (int i=0; i<out_array.length; i++) {
			out_array[i] = out.get(i);
		}
		
		return out_array;
	}
	
	/*
	 * Loads a list of points contained with an area
	 */
	/**
	 * Initializes the loading of data contained within a boundary.
	 * This method returns the number of chunks required for this
	 * particular query. Successive execution of getNextPointList()
	 * returns the array of data.
	 * 
	 * @param area
	 * @return int
	 */
	public int initPointListLoad(Boundary area) {
		selection = area;
		partitionsIndices = queryArea(area);
		currentPartition = 0;
		
		if (DEBUG && partitionsIndices.length > 0) {
			/*System.out.println("Partitions array:");
			for (int i=0; i<partitionsIndices.length-1; i++) {
				System.out.print(partitionsIndices[i] + ", ");
			}
			System.out.println(partitionsIndices[partitionsIndices.length-1]);
			area.printBoundary("partitions query");*/
		}
		
		if (partitionsIndices.length == 0)
			return 1;
		else
			return partitionsIndices.length;
	}
	
	/**
	 * Returns the float values of the image/volume as an ArrayList
	 * 
	 * Points - e.g. [x1, y1, z1, x2, y2, z2, ...]
	 * 
	 * @return ArrayList<Float>
	 */
	public ArrayList<Float> getNextPointList() {
		int ind = partitionsIndices[currentPartition];
		Short[] data = nodes.get(ind).getSingleFile();
		currentPartition ++;
		return PointList.getArrayList(data, nodes.get(ind).b, selection);
	}
	
	/**
	 * Writes a list of points to the current node.
	 * The current node is defined via 'partitionsIndices[currentPartition-1]'.
	 * 
	 * This method is intent to be executed after the getNextPointList() method.
	 * It writes back the modified point list to the file.
	 * 
	 * @param points
	 */
	public void writePointList(ArrayList<Float> points) {
		
		/*System.out.println("new: ");
		System.out.println(partitionsIndices[currentPartition]);
		System.out.println(currentPartition);
		*/
		
		//DEBUG
		/*for (int i=0; i<points.size(); i++) {
			System.out.print(points.get(i) + ", ");
		}
		System.out.println();*/
		
		System.out.println(partitionsIndices.length);
		
		int ind = partitionsIndices[currentPartition-1]; //TODO: counter intuitive! otherwise it wont work
		Short[] data = PointList.getLinearArrayShort(points, nodes.get(ind).b);
		
		//DEBUG
		/*for (int i=0; i<data.length; i++) {
			System.out.print(data[i] + ", ");
		}
		System.out.println();*/
		
		nodes.get(ind).setSingleFile(data);
	}
	
	/**
	 * Returns the size of the data.
	 * 
	 * @return Boundary
	 */
	public Boundary getBoundary() {
		return nodes.get(0).b;
	}
	
	/**
	 * Returns the maximum of the value-range.
	 * [0..maximum]
	 * 
	 * @return float
	 */
	public float getLimit() {
		return (float)Short.MAX_VALUE;
	}
	
	/**
	 * Prints the nodes to the console.
	 */
	public void debug_draw() {
		for (int i=0; i<nodes.size(); i++) {
			System.out.println(i +": "+ nodes.get(i).left_child + "/" + nodes.get(i).right_child);
		}
	}
	
	/**
	 * Prints some of the nodes to the console.
	 */
	public void printTree() {
		System.out.println("0");
		System.out.println(nodes.get(0).left_child + ", " + nodes.get(0).right_child);
		System.out.println(	
				nodes.get(nodes.get(0).left_child).left_child 
				+ ", " + nodes.get(nodes.get(0).left_child).right_child);
		
		nodes.get(2).b.printBoundary("2");
		nodes.get(9).b.printBoundary("9");
		
	}
	
}
